@@ -52,9 +52,7 @@
         (map-pairs stp nsp-tn 0 nl-registers :pre-index -80)
         (inst mov nl0 tmp-tn) ;; size
         #+sb-thread
-        (progn
-          (inst stp cfp-tn csp-tn (@ thread-tn (* thread-control-frame-pointer-slot n-word-bytes)))
-          (inst str csp-tn (@ thread-tn (* thread-foreign-function-call-active-slot n-word-bytes))))
+        (inst stp cfp-tn csp-tn (@ thread-tn (* thread-control-frame-pointer-slot n-word-bytes)))
         #-sb-thread
         (progn
           ;; Each of these loads of a fixup loads the address of a linkage table entry,
@@ -78,8 +76,7 @@
           (inst str (32-bit-reg null-tn) (@ nl1))) ; (alien variable is 4 bytes, not 8)
         ;; Create a new frame
         (inst add csp-tn csp-tn (+ 32 80))
-        (inst stp cfp-tn null-tn (@ csp-tn -112))
-        (inst stp code-tn lr-tn (@ csp-tn -96))
+        (inst stp cfp-tn lr-tn (@ csp-tn -112))
 
         (map-pairs stp csp-tn -80 lisp-registers)
         (map-pairs stp nsp-tn 0 float-registers :pre-index -512 :delta 32)
@@ -90,11 +87,11 @@
         (map-pairs ldp nsp-tn 480 float-registers :post-index 512 :delta -32)
         (map-pairs ldp csp-tn -16 lisp-registers :delta -16)
 
-        (inst ldr lr-tn (@ csp-tn -88))
+        (inst ldr lr-tn (@ csp-tn -104))
 
         (inst sub csp-tn csp-tn (+ 32 80)) ;; deallocate the frame
         #+sb-thread
-        (inst str zr-tn (@ thread-tn (* thread-foreign-function-call-active-slot n-word-bytes)))
+        (inst str zr-tn (@ thread-tn (* thread-control-stack-pointer-slot n-word-bytes)))
         #-sb-thread
         (progn
           ;; We haven't restored NL1 yet, so it's ok to use as a scratch register here.
@@ -111,81 +108,35 @@
   (define-alloc-tramp list-alloc-tramp "alloc_list"))
 
 (define-assembly-routine
-    (xundefined-tramp (:return-style :none)
-                      (:align n-lowtag-bits)
-                      (:export undefined-tramp
-                               (undefined-tramp-tagged
-                                (+ xundefined-tramp
-                                   fun-pointer-lowtag))))
+    (undefined-tramp (:return-style :none))
     ()
-  HEADER
-  (inst dword simple-fun-widetag)
-  (inst dword (make-fixup 'undefined-tramp-tagged
-                          :assembly-routine))
-  (dotimes (i (- simple-fun-insts-offset 2))
-    (inst dword nil-value))
-
-  UNDEFINED-TRAMP
-  (inst adr code-tn header fun-pointer-lowtag)
+  (inst str lr-tn (@ cfp-tn 8))
   (emit-error-break nil cerror-trap (error-number-or-lose 'undefined-fun-error)
                     (list lexenv-tn))
-  (loadw code-tn lexenv-tn closure-fun-slot fun-pointer-lowtag)
-  (inst add lr-tn code-tn (- (* simple-fun-insts-offset n-word-bytes) fun-pointer-lowtag))
-
+  (loadw lr-tn lexenv-tn closure-fun-slot fun-pointer-lowtag)
+  (inst add lr-tn lr-tn 4)
   (inst br lr-tn))
 
 (define-assembly-routine
-    (xundefined-alien-tramp (:return-style :none)
-                      (:align n-lowtag-bits)
-                      (:export undefined-alien-tramp
-                               (undefined-alien-tramp-tagged
-                                (+ xundefined-alien-tramp
-                                   fun-pointer-lowtag))))
-    ((:temp r8-tn unsigned-reg r8-offset))
-  HEADER
-  (inst dword simple-fun-widetag)
-  (inst dword (make-fixup 'undefined-alien-tramp-tagged
-                         :assembly-routine))
-  (dotimes (i (- simple-fun-insts-offset 2))
-    (inst dword nil-value))
-
-  UNDEFINED-ALIEN-TRAMP
-  (inst adr code-tn header fun-pointer-lowtag)
-  (error-call nil 'undefined-alien-fun-error r8-tn))
+    (undefined-alien-tramp (:return-style :none))
+    ((:temp r9-tn unsigned-reg r9-offset))
+  (inst str lr-tn (@ cfp-tn 8))
+  (error-call nil 'undefined-alien-fun-error r9-tn))
 
 (define-assembly-routine
-    (xclosure-tramp (:return-style :none)
-                      (:align n-lowtag-bits)
-                      (:export closure-tramp
-                               (closure-tramp-tagged
-                                (+ xclosure-tramp
-                                   fun-pointer-lowtag))))
+    (closure-tramp (:return-style :none))
     ()
-  (inst dword simple-fun-widetag)
-  (inst dword (make-fixup 'closure-tramp-tagged
-                         :assembly-routine))
-  (dotimes (i (- simple-fun-insts-offset 2))
-    (inst dword nil-value))
-
-  CLOSURE-TRAMP
+  (inst str lr-tn (@ cfp-tn 8))
   (loadw lexenv-tn lexenv-tn fdefn-fun-slot other-pointer-lowtag)
-  (loadw code-tn lexenv-tn closure-fun-slot fun-pointer-lowtag)
-  (inst add lr-tn code-tn (- (* simple-fun-insts-offset n-word-bytes) fun-pointer-lowtag))
+  (loadw lr-tn lexenv-tn closure-fun-slot fun-pointer-lowtag)
+  (inst add lr-tn lr-tn 4)
   (inst br lr-tn))
 
 (define-assembly-routine
-    (xfuncallable-instance-tramp (:return-style :none)
-                      (:align n-lowtag-bits)
-                      (:export (funcallable-instance-tramp
-                                (+ xfuncallable-instance-tramp
-                                   fun-pointer-lowtag))))
+    (funcallable-instance-tramp (:return-style :none))
     ()
-  (inst dword simple-fun-widetag)
-  (inst dword (make-fixup 'funcallable-instance-tramp :assembly-routine))
-  (dotimes (i (- simple-fun-insts-offset 2))
-    (inst dword nil-value))
-
+  (inst str lr-tn (@ cfp-tn 8))
   (loadw lexenv-tn lexenv-tn funcallable-instance-function-slot fun-pointer-lowtag)
-  (loadw code-tn lexenv-tn closure-fun-slot fun-pointer-lowtag)
-  (inst add lr-tn code-tn (- (* simple-fun-insts-offset n-word-bytes) fun-pointer-lowtag))
+  (loadw lr-tn lexenv-tn closure-fun-slot fun-pointer-lowtag)
+  (inst add lr-tn lr-tn 4)
   (inst br lr-tn))
