@@ -32,42 +32,20 @@
     fpr-restore ; KLUDGE: this is element 6 of the entry point vector
     (do-fprs pop :xmm)))
 
-;;; It is arbitrary whether each of the next 4 routines is named ALLOC-something,
-;;; exporting an additional entry named CONS-something, versus being named CONS-something
-;;; and exporting ALLOC-something.  It just depends on which of those you would like
-;;; to have to set and clear the low bit to match ALLOC-DISPATCH.
-;;; Think of "cons" as meaning the generic sense of "consing", not as actually
-;;; allocating conses, because fixed-sized non-cons object allocation may enter
-;;; through the cons entry point. Cons cells *must* use that entry point.
-(define-assembly-routine (alloc->rnn (:export cons->rnn)) ()
-  (inst or :byte (ea 8 rsp-tn) 1)
-  CONS->RNN
+(define-assembly-routine (alloc-tramp) ()
   (with-registers-preserved (c)
     (inst mov rdi-tn (ea 16 rbp-tn))
-    (inst call (make-fixup 'alloc-dispatch :assembly-routine))
+    (inst call (make-fixup "alloc" :foreign))
     (inst mov (ea 16 rbp-tn) rax-tn))) ; result onto stack
 
-(define-assembly-routine (alloc->r11 (:export cons->r11) (:return-style :none)) ()
-  (inst or :byte (ea 8 rsp-tn) 1)
-  CONS->R11
-  (with-registers-preserved (c :except r11-tn)
+(define-assembly-routine (list-alloc-tramp) ()
+  (with-registers-preserved (c)
     (inst mov rdi-tn (ea 16 rbp-tn))
-    (inst call (make-fixup 'alloc-dispatch :assembly-routine))
-    (inst mov r11-tn rax-tn))
-  (inst ret 8)) ; pop argument
-
-(define-assembly-routine (alloc-dispatch (:return-style :none)) ()
-  ;; If RDI has a 0 in the low bit, then we're allocating cons cells.
-  ;; A 1 bit signifies anything other than cons cells, and is equivalent
-  ;; to "could this object consume large-object pages" in gencgc.
-  (inst test :byte rdi-tn 1)
-  (inst jmp :z (make-fixup "alloc_list" :foreign))
-  (inst xor :byte rdi-tn 1) ; clear the bit
-  (inst jmp  (make-fixup "alloc" :foreign)))
+    (inst call (make-fixup "alloc_list" :foreign))
+    (inst mov (ea 16 rbp-tn) rax-tn))) ; result onto stack
 
 ;;; These routines are for the deterministic consing profiler.
 ;;; The C support routine's argument is the return PC.
-;;; FIXME: we're missing routines that preserve YMM. I guess nobody cares.
 (define-assembly-routine (enable-alloc-counter) ()
   (with-registers-preserved (c)
     (inst lea rdi-tn (ea 8 rbp-tn))
@@ -143,22 +121,7 @@
   (inst jmp (object-slot-ea rax-tn closure-fun-slot fun-pointer-lowtag)))
 
 (define-assembly-routine (ensure-symbol-hash (:return-style :raw)) ()
-  (with-registers-preserved (lisp :except r11-tn)
+  (with-registers-preserved (lisp)
     (inst mov rdx-tn (ea 16 rbp-tn)) ; arg
     (call-static-fun 'ensure-symbol-hash 1)
     (inst mov (ea 16 rbp-tn) rdx-tn))) ; result to arg passing loc
-
-(define-assembly-routine (sb-impl::install-hash-table-lock
-                          (:return-style :raw))
-  ()
-  (with-registers-preserved (lisp :except r11-tn)
-    (inst mov rdx-tn (ea 16 rbp-tn)) ; arg
-    (call-static-fun 'sb-impl::install-hash-table-lock 1)
-    (inst mov (ea 16 rbp-tn) rdx-tn))) ; result to arg passing loc
-
-(define-assembly-routine (invalid-layout-trap (:return-style :none)) ()
-  (with-registers-preserved (lisp :except r11-tn)
-    (inst mov rdx-tn (ea 16 rbp-tn)) ; arg
-    (call-static-fun 'update-object-layout 1)
-    (inst mov (ea 16 rbp-tn) rdx-tn)) ; result to arg passing loc
-  (inst ret)) ;  Caller pops the result
