@@ -27,11 +27,10 @@
     (extern-alien "dynamic_space_free_pointer" system-area-pointer)))
 
 (declaim (inline dynamic-usage))
-#+gencgc
-(defun dynamic-usage ()
-  (extern-alien "bytes_allocated" os-vm-size-t))
-#-gencgc
 (defun dynamic-usage ()
+  #+gencgc
+  (extern-alien "bytes_allocated" os-vm-size-t)
+  #-gencgc
   (truly-the word
              (- (sap-int (sb-c::dynamic-space-free-pointer))
                 (current-dynamic-space-start))))
@@ -77,10 +76,20 @@ SB-PROFILE package does), or to design a more microefficient interface
 and submit it as a patch."
   (+ (dynamic-usage)
      *n-bytes-freed-or-purified*))
+
+(defun primitive-object-size (object)
+  "Return number of bytes of heap or stack directly consumed by OBJECT"
+  (cond ((not (sb-vm:is-lisp-pointer (get-lisp-obj-address object))) 0)
+        ((eq object nil) (ash sb-vm::sizeof-nil-in-words sb-vm:word-shift))
+        ((simple-fun-p object) (code-object-size (fun-code-header object)))
+        (t
+         (with-alien ((sizer (function unsigned unsigned) :extern "primitive_object_size"))
+           (with-pinned-objects (object)
+             (alien-funcall sizer (get-lisp-obj-address object)))))))
 
 ;;;; GC hooks
 
-(!define-load-time-global *after-gc-hooks* nil
+(define-load-time-global *after-gc-hooks* nil
   "Called after each garbage collection, except for garbage collections
 triggered during thread exits. In a multithreaded environment these hooks may
 run in any thread.")
@@ -88,8 +97,7 @@ run in any thread.")
 
 ;;;; internal GC
 
-(define-alien-routine collect-garbage int
-  (#+gencgc last-gen #-gencgc ignore int))
+(define-alien-routine collect-garbage int (last-gen int))
 
 #+(or sb-thread sb-safepoint)
 (progn
@@ -100,35 +108,10 @@ run in any thread.")
   (defun gc-stop-the-world ())
   (defun gc-start-the-world ()))
 
-#+gencgc
-(progn
-  (define-alien-variable ("gc_logfile" %gc-logfile) (* char))
-  (defun (setf gc-logfile) (pathname)
-    (let ((new (when pathname
-                 (make-alien-string
-                  (native-namestring (translate-logical-pathname pathname)
-                                     :as-file t))))
-          (old %gc-logfile))
-      (setf %gc-logfile new)
-      (when old
-        (free-alien old))
-      pathname))
-  (defun gc-logfile ()
-    "Return the pathname used to log garbage collections. Can be SETF.
-Default is NIL, meaning collections are not logged. If non-null, the
-designated file is opened before and after each collection, and generation
-statistics are appended to it."
-    (let ((val (cast %gc-logfile c-string)))
-      (when val
-        (native-pathname val)))))
-
 (declaim (inline dynamic-space-size))
 (defun dynamic-space-size ()
   "Size of the dynamic space in bytes."
   (extern-alien "dynamic_space_size" os-vm-size-t))
-#+gencgc
-(define-symbol-macro sb-vm:dynamic-space-end
-  (+ (dynamic-space-size) sb-vm:dynamic-space-start))
 
 ;;;; SUB-GC
 
@@ -243,6 +226,17 @@ statistics are appended to it."
                              (neq epoch *gc-pending*))
                     (return 0)))))))))
 
+#+sb-thread
+(defun post-gc ()
+  (sb-impl::finalizer-thread-notify)
+  (alien-funcall (extern-alien "empty_thread_recyclebin" (function void)))
+  ;; Post-GC actions are invoked synchronously by the GCing thread,
+  ;; which is an arbitrary one. If those actions aquire any locks, or are sensitive
+  ;; to the state of *ALLOW-WITH-INTERRUPTS*, any deadlocks of what-have-you
+  ;; are user error. Hooks need to be sufficiently uncomplicated as to be harmless.
+  (call-hooks "after-GC" *after-gc-hooks* :on-error :warn))
+
+#-sb-thread
 (defun post-gc ()
   ;; Outside the mutex, interrupts may be enabled: these may cause
   ;; another GC. FIXME: it can potentially exceed maximum interrupt
@@ -272,18 +266,7 @@ statistics are appended to it."
   ;; but it's not permissible to invoke CONDITION-NOTIFY from a
   ;; dying thread, so we still need the guard for that, but not
   ;; the guard for whether interupts are enabled.
-  #+sb-thread (alien-funcall (extern-alien "empty_thread_recyclebin" (function void)))
-  (let ((threadp #+sb-thread (%instancep sb-impl::*finalizer-thread*)))
-    (when threadp
-      ;; grabbing a lock, need to clear out any waiting-for to avoid
-      ;; spurios deadlock detections.
-      (sb-thread::without-thread-waiting-for ()
-        ;; It's OK to frob a condition variable regardless of
-        ;; *allow-with-interrupts*, and probably OK to start a thread.
-        ;; For consistency with the previous behavior, we delay finalization
-        ;; if there is no finalizer thread and interrupts are disabled.
-        ;; That's my excuse anyway, not having looked more in-depth.
-        (run-pending-finalizers)))
+
     ;; Here's one reason that MAX_INTERRUPTS is as high as it is.
     ;; If the thread that performed GC runs post-GC hooks which cons enough to
     ;; cause another GC while in the hooks, then as soon as interrupts are allowed
@@ -294,40 +277,19 @@ statistics are appended to it."
     ;; pseudo-atomic sequence to take a signal hit. So with interrupts eanbled,
     ;; we get back into the GC, which calls post-GC, which might cons ...
     ;; See the example at the bottom of src/code/final for a clear picture.
-    ;;
-    ;; To mitigate that problem, if you have no hooks, we can avoid inducing overflow
-    ;; of the interrupt contexts. Relegating post-GC actions to their own thread
-    ;; (as suggested above) would largely solve this - the post-GC action would be to
-    ;; just "kick" that thread to do its thing. If it doesn't finish in time (by the
-    ;; time we kick it again), that's its problem, not the GC's problem.
-    ;;
-    ;; If we do implement a general post-GC thread, there remains a question of
-    ;; of what to do with post-GC actions for #-sb-thread that cons too much,
-    ;; risking recursive invocation of GC. Pretty much don't do that.
-    ;;
-    (when (and *allow-with-interrupts*
-               ;; Continue if any of the following:
-               ;; - you want a finalizer thread but it hasn't been started
-               ;; - you don't want a finalizer thread, and we're not already
-               ;;   in SCAN-FINALIZERS
-               ;; - there are some hooks to run
-               ;; Logically, the finalizer existence test belongs in src/code/final,
-               ;; but the flaw in that is we're not going to call that code until unmasking
-               ;; interrupts, which is precisely the thing we need to NOT do if already
-               ;; in post-GC code of any kind (be it finalizer or other).
-               (or #+sb-thread (if sb-impl::*finalizer-thread*
-                                   (not threadp)
-                                   (not sb-impl::*in-a-finalizer*))
-                   #-sb-thread (and (sb-impl::hash-table-culled-values
-                                     (sb-impl::finalizer-id-map
-                                      sb-impl::**finalizer-store**))
-                                    (not sb-impl::*in-a-finalizer*))
+    ;; Logically, the finalizer existence test belongs in src/code/final,
+    ;; but the flaw in that is we're not going to call that code until unmasking
+    ;; interrupts, which is precisely the thing we need to NOT do if already
+    ;; in post-GC code of any kind (be it finalizer or other).
+  (when (and *allow-with-interrupts*
+               (or (and (sb-impl::hash-table-culled-values
+                         (sb-impl::finalizer-id-map sb-impl::**finalizer-store**))
+                        (not sb-impl::*in-a-finalizer*))
                    *after-gc-hooks*))
       (sb-thread::without-thread-waiting-for ()
         (with-interrupts
-          (unless threadp
-            (run-pending-finalizers))
-          (call-hooks "after-GC" *after-gc-hooks* :on-error :warn))))))
+          (run-pending-finalizers)
+          (call-hooks "after-GC" *after-gc-hooks* :on-error :warn)))))
 
 ;;; This is the user-advertised garbage collection function.
 (defun gc (&key (full nil) (gen 0) &allow-other-keys)
@@ -395,7 +357,7 @@ Note: currently changes to this value are lost when saving core."
   (extern-alien "bytes_consed_between_gcs" os-vm-size-t))
 
 (defun (setf bytes-consed-between-gcs) (val)
-  (declare (type index val))
+  (declare (type (and fixnum unsigned-byte) val))
   (setf (extern-alien "bytes_consed_between_gcs" os-vm-size-t)
         val))
 
@@ -408,16 +370,44 @@ Note: currently changes to this value are lost when saving core."
 
 ;;;; GENCGC specifics
 ;;;;
-;;;; For documentation convenience, these have stubs on non-GENCGC platforms
-;;;; as well.
 #+gencgc
+(progn
+
+;; A small amount of backward-compatibility in case anyone uses
+;; the -CARD- size. Of couse, it's about to become different
+;; from the -PAGE- size, so any code that uses this is probably wrong.
+(define-symbol-macro sb-vm:gencgc-card-bytes sb-vm:gencgc-page-bytes)
+
+(define-symbol-macro sb-vm:dynamic-space-end
+    (+ (dynamic-space-size) sb-vm:dynamic-space-start))
+
+(define-alien-variable ("gc_logfile" %gc-logfile) (* char))
+
+(defun (setf gc-logfile) (pathname)
+  (let ((new (when pathname
+               (make-alien-string
+                (native-namestring (translate-logical-pathname pathname)
+                                   :as-file t))))
+        (old %gc-logfile))
+    (setf %gc-logfile new)
+    (when old
+      (free-alien old))
+    pathname))
+(defun gc-logfile ()
+    "Return the pathname used to log garbage collections. Can be SETF.
+Default is NIL, meaning collections are not logged. If non-null, the
+designated file is opened before and after each collection, and generation
+statistics are appended to it."
+  (let ((val (cast %gc-logfile c-string)))
+    (when val
+      (native-pathname val))))
+
 (deftype generation-index ()
   `(integer 0 ,sb-vm:+pseudo-static-generation+))
 
 ;;; FIXME: GENERATION (and PAGE, as seen in room.lisp) should probably be
 ;;; defined in Lisp, and written to header files by genesis, instead of this
 ;;; OAOOMiness -- this duplicates the struct definition in gencgc.c.
-#+gencgc
 (define-alien-type generation
     (struct generation
             (bytes-allocated os-vm-size-t)
@@ -428,34 +418,43 @@ Note: currently changes to this value are lost when saving core."
             (cum-sum-bytes-allocated os-vm-size-t)
             (minimum-age-before-gc double)))
 
-#+gencgc
 (define-alien-variable generations
     (array generation #.(1+ sb-vm:+pseudo-static-generation+)))
 
-(macrolet ((def (slot doc &optional setfp)
+(export 'page-protected-p)
+(macrolet ((addr->mark (addr)
+             `(sap-ref-8 (extern-alien "gc_card_mark" system-area-pointer)
+                         (logand (ash ,addr (- (integer-length (1- sb-vm:gencgc-card-bytes))))
+                                 (extern-alien "gc_card_table_mask" int)))))
+  (defun page-protected-p (object) ; OBJECT must be pinned by caller
+    (eql #xff (addr->mark (get-lisp-obj-address object))))
+  (defun object-card-marks (obj)
+    (aver (eq (heap-allocated-p obj) :dynamic))
+    ;; Return a bit-vector with a 1 for each marked card the object is on
+    ;; and a 0 for each unmarked card.
+    (with-pinned-objects (obj)
+      (let* ((base (logandc2 (get-lisp-obj-address obj) sb-vm:lowtag-mask))
+             (last (+ base (1- (primitive-object-size obj))))
+             (aligned-base (logandc2 base (1- sb-vm:gencgc-card-bytes)))
+             (aligned-last (logandc2 last (1- sb-vm:gencgc-card-bytes)))
+             (result (make-array (1+ (truncate (- aligned-last aligned-base)
+                                               sb-vm:gencgc-card-bytes))
+                                 :element-type 'bit)))
+        (loop for addr from aligned-base to aligned-last by sb-vm:gencgc-card-bytes
+              for i from 0
+              do (setf (aref result i) (if (eql (addr->mark addr) 0) 1 0)))
+        result))))
+
+(macrolet ((def (slot doc &optional adjustable)
              `(progn
                 (defun ,(symbolicate "GENERATION-" slot) (generation)
                   ,doc
-                  #+gencgc
                   (declare (generation-index generation))
-                  #-gencgc
-                  (declare (ignore generation))
-                  #-gencgc
-                  (error "~S is a GENCGC only function and unavailable in this build"
-                         ',slot)
-                  #+gencgc
                   (slot (deref generations generation) ',slot))
-                ,@(when setfp
-                        `((defun (setf ,(symbolicate "GENERATION-" slot)) (value generation)
-                            #+gencgc
-                            (declare (generation-index generation))
-                            #-gencgc
-                            (declare (ignore value generation))
-                            #-gencgc
-                            (error "(SETF ~S) is a GENCGC only function and unavailable in this build"
-                                   ',slot)
-                            #+gencgc
-                            (setf (slot (deref generations generation) ',slot) value)))))))
+                ,@(when adjustable
+                    `((defun (setf ,(symbolicate "GENERATION-" slot)) (value generation)
+                        (declare (generation-index generation))
+                        (setf (slot (deref generations generation) ',slot) value)))))))
   (def bytes-consed-between-gcs
       "Number of bytes that can be allocated to GENERATION before that
 generation is considered for garbage collection. This value is meaningless for
@@ -496,16 +495,11 @@ objects allocated to the generation have seen younger objects promoted to it.
 Available on GENCGC platforms only.
 
 Experimental: interface subject to change."
-    #+gencgc
     (declare (generation-index generation))
-    #-gencgc (declare (ignore generation))
-    #-gencgc
-    (error "~S is a GENCGC only function and unavailable in this build."
-           'generation-average-age)
-    #+gencgc
     (alien-funcall (extern-alien "generation_average_age"
                                  (function double generation-index-t))
                    generation))
+)
 
 (macrolet ((cases ()
              `(cond ((< (current-dynamic-space-start) addr
