@@ -98,9 +98,11 @@
 (defknown find-classoid (name-for-class &optional t)
   (or classoid null) ())
 (defknown classoid-of (t) classoid (flushable))
-(defknown layout-of (t) layout (flushable))
-#+64-bit (defknown layout-depthoid (layout) fixnum (flushable always-translatable))
-(defknown layout-depthoid-gt (layout integer) boolean (flushable))
+(defknown wrapper-of (t) wrapper (flushable))
+(defknown wrapper-depthoid (wrapper) layout-depthoid (flushable))
+#+64-bit (defknown layout-depthoid (sb-vm:layout) layout-depthoid (flushable always-translatable))
+#+(or x86 x86-64) (defknown (layout-depthoid-ge) (sb-vm:layout integer) boolean (flushable))
+(defknown %structure-is-a (instance t) boolean (foldable flushable))
 (defknown copy-structure (structure-object) structure-object
   (flushable)
   :derive-type #'result-type-first-arg)
@@ -122,7 +124,7 @@
 (defknown (symbol-function) (symbol) function ())
 
 (defknown boundp (symbol) boolean (flushable))
-(defknown fboundp ((or symbol cons)) boolean (unsafely-flushable))
+(defknown fboundp ((or symbol cons)) (or null function) (unsafely-flushable))
 (defknown special-operator-p (symbol) t
   ;; The set of special operators never changes.
   (movable foldable flushable))
@@ -341,9 +343,15 @@
 (defknown %multiply-high (word word) word
     (movable foldable flushable))
 
-(defknown (mod rem) (real real) real
+(defknown multiply-fixnums (fixnum fixnum) integer
   (movable foldable flushable))
 
+(defknown %signed-multiply-high (sb-vm:signed-word sb-vm:signed-word) sb-vm:signed-word
+    (movable foldable flushable))
+
+(defknown (mod rem) (real real) real
+    (movable foldable flushable))
+
 (defknown (ffloor fceiling fround ftruncate)
   (real &optional real) (values float real)
   (movable foldable flushable))
@@ -362,7 +370,18 @@
 ;;; e.g. the behavior if the first arg is a NaN is well-defined as we have it,
 ;;; but what about the second arg? We need some test cases around this.
 (defknown float-sign (float &optional float) float
-  (movable foldable unsafely-flushable))
+  (movable foldable unsafely-flushable)
+  :derive-type (lambda (call &aux (args (combination-args call))
+                                  (type (unless (cdr args) (lvar-type (first args)))))
+                 (cond ((and type (csubtypep type (specifier-type 'single-float)))
+                        (specifier-type '(member $1f0 $-1f0)))
+                       ((and type (csubtypep type (specifier-type 'double-float)))
+                        (specifier-type '(member $1d0 $-1d0)))
+                       (type
+                        (specifier-type '(member $1f0 $-1f0 $1d0 $-1d0)))
+                       (t
+                        (specifier-type 'float)))))
+
 (defknown (float-digits float-precision) (float) float-digits
   (movable foldable unsafely-flushable))
 (defknown integer-decode-float (float)
@@ -577,6 +596,19 @@
     ()
   :derive-type #'result-type-first-arg
   :result-arg 0)
+;;; Like FILL but with no keyword argument parsing
+(defknown quickfill ((modifying (simple-array * 1)) t) (simple-array * 1) ()
+  :derive-type #'result-type-first-arg
+  :result-arg 0)
+;;; Special case of FILL that takes either a machine word with which to fill,
+;;; or a keyword indicating a certain behavior to compute the word.
+;;; In either case the supplied count is lispwords, not elements.
+;;; This might be a no-op depending on whether memory is prezeroized.
+(defknown sb-vm::splat ((modifying (simple-array * 1)) index (or symbol sb-vm:word))
+    (simple-array * 1)
+  (always-translatable)
+  :derive-type #'result-type-first-arg
+  :result-arg 0)
 
 (defknown replace ((modifying sequence) proper-sequence &rest t &key (:start1 index)
                    (:end1 sequence-end) (:start2 index) (:end2 sequence-end))
@@ -875,22 +907,20 @@
 (defknown list* (t &rest t) t (movable flushable))
 ;;; The length constraint on MAKE-LIST is such that:
 ;;; - not every byte of addressable memory can be used up.
-;;; - the number of bytes to allocate should not be a bignum,
-;;    nor can its native representation be larger than 'sword_t'
-;;; So it's half of most-positive-word divided by the cons size,
-;;; which is roughly twice as small as array-dimension-limit on 32-bit machines,
-;;; or 8x smaller on 64-bit machines.
+;;; - the number of bytes to allocate should be a fixnum
+;;; - type-checking can use use efficient bit-masking approach
+;;;   to combine the fixnum + range test into one instruction
 (eval-when (:compile-toplevel :load-toplevel :execute)
   (defconstant make-list-limit
-    (floor (/ sb-ext:most-positive-word 2) (* 2 sb-vm:n-word-bytes))))
-(defknown make-list ((integer 0 (#.make-list-limit)) &key (:initial-element t)) list
+    (ash most-positive-fixnum (- (+ sb-vm:word-shift 1)))))
+(defknown make-list ((integer 0 #.make-list-limit) &key (:initial-element t)) list
   (movable flushable))
-(defknown %make-list ((integer 0 (#.make-list-limit)) t) list (movable flushable))
+(defknown %make-list ((integer 0 #.make-list-limit) t) list (movable flushable))
 
-(defknown sb-impl::|List| (&rest t) list (movable flushable foldable))
-(defknown sb-impl::|List*| (t &rest t) t (movable flushable foldable))
-(defknown sb-impl::|Append| (&rest t) t (flushable foldable))
-(defknown sb-impl::|Vector| (&rest t) simple-vector (flushable foldable))
+(defknown sb-impl::|List| (&rest t) list (movable flushable))
+(defknown sb-impl::|List*| (t &rest t) t (movable flushable))
+(defknown sb-impl::|Append| (&rest t) t (movable flushable))
+(defknown sb-impl::|Vector| (&rest t) simple-vector (movable flushable))
 
 ;;; All but last must be of type LIST, but there seems to be no way to
 ;;; express that in this syntax.
@@ -1032,6 +1062,7 @@
         (:synchronized t))
   hash-table
   (flushable))
+(defknown sb-impl::make-hash-table-using-defaults (integer) hash-table (flushable))
 (defknown hash-table-p (t) boolean (movable foldable flushable))
 (defknown gethash (t hash-table &optional t) (values t boolean)
   (flushable)) ; not FOLDABLE, since hash table contents can change
@@ -1082,7 +1113,7 @@
                        (:displaced-to (or array null))
                        (:displaced-index-offset index))
     array (flushable))
-
+(defknown initial-contents-error (t t) nil)
 (defknown fill-data-vector (vector list sequence) vector ()
   :result-arg 0)
 
@@ -1154,7 +1185,10 @@
   (movable foldable flushable))
 (defknown fill-pointer (complex-vector) index
     (unsafely-flushable))
-(defknown sb-impl::fill-pointer-error (t &optional t) nil)
+;;; We could cut 1 instruction off the call sequence by not loading the arg count
+;;; register for error-handler helpers that take fixed args.
+;;; Perhaps they should all be trap instead though?
+(defknown sb-vm::fill-pointer-error (t) nil)
 
 (defknown vector-push (t (modifying complex-vector)) (or index null) ())
 (defknown vector-push-extend (t (modifying complex-vector) &optional (and index (integer 1))) index
@@ -1162,15 +1196,15 @@
 (defknown vector-pop ((modifying complex-vector)) t ())
 
 ;;; FIXME: complicated MODIFYING
-;;; Also, an important-result warning could be provided if the array
-;;; is known to be not expressly adjustable.
 (defknown adjust-array
   (array (or index list) &key (:element-type type-specifier)
          (:initial-element t) (:initial-contents t)
          (:fill-pointer (or index boolean))
          (:displaced-to (or array null))
          (:displaced-index-offset index))
-  array ())
+  ;; This is a special case in CHECK-IMPORTANT-RESULT because it is not
+  ;; necessary to use the result if the array is adjustable.
+  array (important-result))
 ;  :derive-type 'result-type-arg1) Not even close...
 
 ;;;; from the "Strings" chapter:
@@ -1646,21 +1680,29 @@
    &key
 
    ;; ANSI options
-   (:output-file (or pathname-designator
-                     null
-                     ;; FIXME: This last case is a non-ANSI hack.
-                     (member t)))
+   (:output-file pathname-designator)
    (:verbose t)
    (:print t)
    (:external-format external-format-designator)
-   (:progress t)
 
    ;; extensions
+   (:progress t)
    (:trace-file t)
    (:block-compile t)
    (:entry-points list)
    (:emit-cfasl t))
   (values (or pathname null) boolean boolean))
+(defknown sb-c::compile-files
+    (cons &key (:output-file pathname-designator)
+               (:verbose t)
+               (:print t)
+               (:external-format external-format-designator)
+               (:progress t)
+               (:trace-file t)
+               (:block-compile t)
+               (:entry-points list)
+               (:emit-cfasl t))
+  (values (or pathname null) boolean boolean))
 
 (defknown (compile-file-pathname)
   (pathname-designator &key (:output-file (or pathname-designator
@@ -1675,7 +1717,7 @@
                        &key (:stream stream) (:use-labels t))
   null)
 
-(defknown describe (t &optional (or stream (member t nil))) (values))
+(defknown describe (t &optional stream-designator) (values))
 (defknown function-lambda-expression (function) (values t boolean t))
 (defknown inspect (t) (values))
 (defknown room (&optional (member t nil :default)) (values))
@@ -1763,7 +1805,7 @@
 ;;; We can't fold this in general because of SATISFIES. There is a
 ;;; special optimizer anyway.
 (defknown %typep (t (or type-specifier ctype)) boolean (movable flushable))
-(defknown %instance-typep (t (or type-specifier ctype layout)) boolean
+(defknown %instance-typep (t (or type-specifier ctype wrapper)) boolean
   (movable flushable always-translatable))
 ;;; We should never emit a call to %typep-wrapper
 (defknown %typep-wrapper (t t (or type-specifier ctype)) t
@@ -1781,7 +1823,8 @@
 (defknown %special-unbind (&rest symbol) t)
 (defknown %listify-rest-args (t index) list (flushable))
 (defknown %more-arg-context (t t) (values t index) (flushable))
-(defknown %more-arg (t index) t)
+(defknown %more-arg (t index) t (flushable))
+(defknown %more-keyword-pair (t fixnum) (values t t) (flushable))
 #+stack-grows-downward-not-upward
 ;;; FIXME: The second argument here should really be NEGATIVE-INDEX, but doing that
 ;;; breaks the build, and I cannot seem to figure out why. --NS 2006-06-29
@@ -1804,11 +1847,10 @@
 (defknown %pop-values (t) t)
 (defknown %nip-values (t t &rest t) (values))
 (defknown %dummy-dx-alloc (t t) t)
-(defknown %allocate-closures (t) *)
 (defknown %type-check-error (t t t) nil)
 (defknown %type-check-error/c (t t t) nil)
 
-;; FIXME: This function does not return, but due to the implementation
+;; %compile-time-type-error does not return, but due to the implementation
 ;; of FILTER-LVAR we cannot write it here.
 (defknown (%compile-time-type-error %compile-time-type-style-warn) (t t t t t t) *)
 (defknown (etypecase-failure ecase-failure) (t t) nil)
@@ -1823,14 +1865,15 @@
 (defknown (%check-bound check-bound) (array index t) index
   (dx-safe))
 (defknown data-vector-ref (simple-array index) t
-  (foldable unsafely-flushable always-translatable))
+  (foldable flushable always-translatable))
 (defknown data-vector-ref-with-offset (simple-array fixnum fixnum) t
-  (foldable unsafely-flushable always-translatable))
+  (foldable flushable always-translatable))
 (defknown data-nil-vector-ref (simple-array index) nil
   (always-translatable))
-(defknown data-vector-set (array index t) t
-  (always-translatable))
-(defknown data-vector-set-with-offset (array fixnum fixnum t) t
+;;; The lowest-level vector SET operators should not return a value.
+;;; Functions built upon them may return the input value.
+(defknown data-vector-set (array index t) (values) (always-translatable))
+(defknown data-vector-set-with-offset (array fixnum fixnum t) (values)
   (always-translatable))
 (defknown hairy-data-vector-ref (array index) t (foldable))
 (defknown hairy-data-vector-set (array index t) t ())
@@ -1947,8 +1990,11 @@
 (defknown %scharset ((modifying simple-string) index character) character ())
 (defknown %set-symbol-value (symbol t) t ())
 (defknown (setf symbol-function) (function symbol) function ())
-(defknown %set-symbol-plist (symbol list) list ()
-  :derive-type #'result-type-last-arg)
+;; Does this really need a type deriver? It's inline, and returns its 1st arg,
+;; i.e. we know exactly what object it returns, which is more precise than
+;; just knowing the type.
+(defknown (setf symbol-plist) (list symbol) list ()
+  :derive-type #'result-type-first-arg)
 (defknown %setnth (unsigned-byte (modifying list) t) t ()
   :derive-type #'result-type-last-arg)
 (defknown %set-fill-pointer ((modifying complex-vector) index) index ()
@@ -1956,7 +2002,7 @@
 
 ;;;; ALIEN and call-out-to-C stuff
 
-(defknown %alien-funcall ((or string system-area-pointer) alien-type &rest *) *)
+(defknown %alien-funcall ((or string system-area-pointer) alien-type &rest t) *)
 
 ;; Used by WITH-PINNED-OBJECTS
 #+(or x86 x86-64)
@@ -2001,7 +2047,7 @@
 ;;; Unfortunately it prints that noise at each call site.
 ;;; Using DEFKNOWN we can make the proclamation effective when
 ;;; running the cross-compiler but not when building it.
-;;; Alternatively we could use SB-XC:PROCLAIM, except that that
+;;; Alternatively we could use PROCLAIM, except that that
 ;;; doesn't exist soon enough, and would need conditionals guarding
 ;;; it, which is sort of the very thing this is trying to avoid.
 (defknown missing-arg () nil)
@@ -2038,6 +2084,10 @@
 (defknown sb-kernel::gc-safepoint () (values) ())
 
 ;;;; atomic ops
+;;; the CAS functions are transformed to something else rather than "translated".
+;;; either way, they should not be called.
+(defknown (cas svref) (t t simple-vector index) t (always-translatable))
+(defknown (cas symbol-value) (t t symbol) t (always-translatable))
 (defknown %compare-and-swap-svref (simple-vector index t t) t
     ())
 (defknown (%compare-and-swap-symbol-value
@@ -2056,7 +2106,7 @@
 ;;; Avoid a ton of FBOUNDP checks in the string stream constructors etc,
 ;;; by wiring in the needed functions instead of dereferencing their fdefns.
 (defknown (ill-in ill-bin ill-out ill-bout
-           sb-impl::string-inch sb-impl::string-in-misc
+           sb-impl::string-in-misc
            sb-impl::string-sout
            sb-impl::finite-base-string-ouch sb-impl::finite-base-string-out-misc
            sb-impl::fill-pointer-ouch sb-impl::fill-pointer-sout
@@ -2104,3 +2154,12 @@
            sb-thread::call-with-system-mutex/without-gcing
            sb-thread::call-with-recursive-system-lock)
     (function t) *))
+
+#+round-float
+(progn
+  (defknown round-double (double-float #1=(member :round :floor :ceiling :truncate))
+      double-float
+      (foldable flushable movable always-translatable))
+
+  (defknown round-single (single-float #1#) single-float
+      (foldable flushable movable always-translatable)))
