@@ -123,7 +123,15 @@
 (defknown %set-symbol-hash (symbol hash-code)
   t ())
 
-(defknown symbol-info-vector (symbol) (or null simple-vector))
+;;; SYMBOL-PACKAGE-ID for #+compact-symbol demands a vop which avoids loading
+;;; a raw bit value in a descriptor register (the SLOT vop returns a descriptor)
+(defknown sb-impl::symbol-package-id (symbol) (unsigned-byte 16))
+;;; TODO: I'd like to eliminate the (OR NULL) from this return type.
+;;; For that to happen, I probably need +nil-packed-infos+ to become
+;;; placed in static space because assembly routines may need it.
+;;; On the other hand, they may not, because there is no special case
+;;; code needed when reading from it, which is entire point.
+(defknown symbol-dbinfo (symbol) (or null packed-info))
 
 (defknown initialize-vector ((simple-array * (*)) &rest t)
   (simple-array * (*))
@@ -158,19 +166,20 @@
 (defknown (function-header-word) (function) sb-vm:word (flushable))
 (defknown (instance-header-word) (instance) sb-vm:word (flushable))
 
-;;; This unconventional setter returns its first arg, not the newval.
 (defknown set-header-data
-    (t (unsigned-byte #.(- sb-vm:n-word-bits sb-vm:n-widetag-bits))) t)
+    (t (unsigned-byte #.(- sb-vm:n-word-bits sb-vm:n-widetag-bits))) (values))
 ;;; Like SET-HEADER-DATA, but instead of writing the entire header,
 ;;; LOGIOR of the specified value into the "data" portion of the word.
+;;; Returns the first argument, *not* the modified header data.
 (defknown logior-header-bits (t (unsigned-byte 16)) t
     (#+x86-64 always-translatable))
 ;;; ASSIGN-VECTOR-FLAGSS assign all and only the flags byte.
 ;;; RESET- performs LOGANDC2 and returns no value.
 (defknown (assign-vector-flags reset-header-bits)
-  (t (unsigned-byte 8)) (values)
+  (t (unsigned-byte 16)) (values)
   (#+x86-64 always-translatable))
-(defknown (test-header-bit)
+;;; test bits of "HeaderData" which start 8 bits over from the lsb
+(defknown (test-header-data-bit)
   (t (unsigned-byte #.(- sb-vm:n-word-bits sb-vm:n-widetag-bits))) (boolean)
   (flushable))
 
@@ -181,10 +190,12 @@
 (defknown %array-rank (array) array-rank
   (flushable))
 
-#+(or x86 x86-64)
+#+(or x86 x86-64 arm64)
 (defknown (%array-rank= widetag=) (t t) boolean
   (flushable))
 
+(defknown simple-array-header-of-rank-p (t array-rank) boolean
+  (flushable))
 (defknown sb-kernel::check-array-shape (simple-array list)
   (simple-array)
   (flushable)
@@ -208,7 +219,6 @@
   (foldable flushable))
 (defknown %set-instance-layout (instance sb-vm:layout) (values) ())
 ;;; %SET-FUN-LAYOUT should only called on FUNCALLABLE-INSTANCE
-;;; (but %set-funcallable-instance-layout is too long a name)
 (defknown %set-fun-layout (funcallable-instance sb-vm:layout) (values) ())
 ;;; Layout getter that accepts any object, and if it has INSTANCE- or FUN-
 ;;; POINTER-LOWTAG returns the layout, otherwise some agreed-upon layout.
@@ -231,6 +241,7 @@
 #+riscv
 (defknown %raw-instance-cas/signed-word (instance index sb-vm:signed-word sb-vm:signed-word)
   sb-vm:signed-word ())
+(defknown %raw-instance-xchg/word (instance index sb-vm:word) sb-vm:word ())
 
 #.`(progn
      ,@(map 'list
@@ -255,15 +266,13 @@
 ;;; These two are mostly used for bit-bashing operations.
 (defknown %vector-raw-bits (t index) sb-vm:word
   (flushable))
-(defknown (%set-vector-raw-bits) (t index sb-vm:word) sb-vm:word
-  ())
+(defknown (%set-vector-raw-bits) (t index sb-vm:word) (values) ())
 
 
 ;;; Allocate an unboxed, non-fancy vector with type code TYPE, length LENGTH,
 ;;; and WORDS words long. Note: it is your responsibility to ensure that the
 ;;; relation between LENGTH and WORDS is correct.
-;;; The extra bit beyond N_WIDETAG_BITS is for the vector weakness flag.
-;;; Note that in almost all situations the first argument is a constant.
+;;; Note that in almost all situations the first argument (TYPE) is a constant.
 ;;; There are only 3 places that it can be non-constant, and not at all
 ;;; after self-build is complete. The three non-constant places are from:
 ;;;  - ALLOCATE-VECTOR-WITH-WIDETAG in src/code/array
@@ -273,7 +282,7 @@
 ;;; vector type will usuallly end up calling allocate-vector-with-widetag
 ;;; via %MAKE-ARRAY.
 (defknown allocate-vector (#+ubsan boolean
-                           (unsigned-byte 9) index
+                           word index
                            ;; The number of words is later converted
                            ;; to bytes, make sure it fits.
                            (and index
@@ -329,10 +338,10 @@
   (defknown %make-simd-pack-ub32 ((unsigned-byte 32) (unsigned-byte 32)
                                   (unsigned-byte 32) (unsigned-byte 32))
       (simd-pack integer)
-      (flushable movable foldable))
+      (flushable movable #-sb-xc-host foldable))
   (defknown %make-simd-pack-ub64 ((unsigned-byte 64) (unsigned-byte 64))
       (simd-pack integer)
-      (flushable movable foldable))
+      (flushable movable #-sb-xc-host foldable))
   (defknown (%simd-pack-low %simd-pack-high) (simd-pack)
       (unsigned-byte 64)
       (flushable movable foldable))
@@ -442,60 +451,90 @@
 ;;;; bignum operations
 
 (defknown %allocate-bignum (bignum-length) bignum
-  (flushable))
+  (flushable #-bignum-assertions always-translatable))
 
 (defknown %bignum-length (bignum) bignum-length
-  (foldable flushable))
+  (foldable flushable always-translatable))
 
-(defknown %bignum-set-length (bignum bignum-length) bignum
-  ())
+;;; Change the length of bignum to be newlen. Newlen must be the same or
+;;; smaller than the old length, and any elements beyond newlen must be zeroed.
+(defknown %bignum-set-length (bignum bignum-length) (values)
+  (always-translatable))
 
 (defknown %bignum-ref (bignum bignum-index) bignum-element-type
-  (flushable))
+  (flushable #-bignum-assertions always-translatable))
 #+(or x86 x86-64)
 (defknown %bignum-ref-with-offset (bignum fixnum (signed-byte 24))
   bignum-element-type (flushable always-translatable))
 
-(defknown %bignum-set (bignum bignum-index bignum-element-type) (values) ())
+(defknown (%bignum-set #+bignum-assertions %%bignum-set) (bignum bignum-index bignum-element-type)
+  (values)
+  (#-bignum-assertions always-translatable))
 
+;;; Return T if digit is positive, or NIL if negative.
 (defknown %digit-0-or-plusp (bignum-element-type) boolean
-  (foldable flushable movable))
+  (foldable flushable movable always-translatable))
 
+;;; %ADD-WITH-CARRY returns a bignum digit and a carry resulting from adding
+;;; together a, b, and an incoming carry.
+;;; %SUBTRACT-WITH-BOROW returns a bignum digit and a borrow resulting from
+;;; subtracting b from a, and subtracting a possible incoming borrow.
 (defknown (%add-with-carry %subtract-with-borrow)
           (bignum-element-type bignum-element-type (mod 2))
   (values bignum-element-type (mod 2))
-  (foldable flushable movable))
+  (foldable flushable movable always-translatable))
 
+;;; This multiplies x-digit and y-digit, producing high and low digits
+;;; manifesting the result. Then it adds the low digit, res-digit, and
+;;; carry-in-digit. Any carries (note, you still have to add two digits
+;;; at a time possibly producing two carries) from adding these three
+;;; digits get added to the high digit from the multiply, producing the
+;;; next carry digit.  Res-digit is optional since two uses of this
+;;; primitive multiplies a single digit bignum by a multiple digit
+;;; bignum, and in this situation there is no need for a result buffer
+;;; accumulating partial results which is where the res-digit comes
+;;; from.
 (defknown %multiply-and-add
           (bignum-element-type bignum-element-type bignum-element-type
                                &optional bignum-element-type)
   (values bignum-element-type bignum-element-type)
-  (foldable flushable movable))
+  (foldable flushable movable always-translatable))
 
+;;; Multiply two digit-size numbers, returning a 2*digit-size result
+;;; split into two digit-size quantities.
 (defknown %multiply (bignum-element-type bignum-element-type)
   (values bignum-element-type bignum-element-type)
-  (foldable flushable movable))
+  (foldable flushable movable always-translatable))
 
 (defknown %lognot (bignum-element-type) bignum-element-type
-  (foldable flushable movable))
+  (foldable flushable movable always-translatable))
 
 (defknown (%logand %logior %logxor) (bignum-element-type bignum-element-type)
   bignum-element-type
   (foldable flushable movable))
 
 (defknown %fixnum-to-digit (fixnum) bignum-element-type
-  (foldable flushable movable))
+  (foldable flushable movable #-(or arm arm64) always-translatable))
 
+;;; This takes three digits and returns the FLOOR'ed result of
+;;; dividing the first two as a 2*digit-size integer by the third.
 (defknown %bigfloor (bignum-element-type bignum-element-type bignum-element-type)
   (values bignum-element-type bignum-element-type)
-  (foldable flushable movable))
+  (foldable flushable movable always-translatable))
 
+;;; Convert the input, a BIGNUM-ELEMENT-TYPE, to a signed word.
+;;; FIXME: considering that both the input and output have N-WORD-BITS significant bits,
+;;; this is really a bad name for the operation.
 (defknown %fixnum-digit-with-correct-sign (bignum-element-type) sb-vm:signed-word
-  (foldable flushable movable))
+  (foldable flushable movable always-translatable))
 
+;;; %ASHR- take a digit-size quantity and shift it to the left,
+;;; returning a digit-size quantity.
+;;; %ASHR- Do an arithmetic shift right of data even though bignum-element-type is
+;;; unsigned.
 (defknown (%ashl %ashr %digit-logical-shift-right)
           (bignum-element-type (mod #.sb-vm:n-word-bits)) bignum-element-type
-  (foldable flushable movable))
+  (foldable flushable movable always-translatable))
 
 ;;;; bit-bashing routines
 
@@ -553,17 +592,18 @@
 (defknown fdefn-name (fdefn) t (foldable flushable))
 (defknown fdefn-fun (fdefn) (or function null) (flushable))
 (defknown (setf fdefn-fun) (function fdefn) t ())
-(defknown fdefn-makunbound (fdefn) t ())
+(defknown fdefn-makunbound (fdefn) (values) ())
 ;;; FDEFN -> FUNCTION, trapping if not FBOUNDP
 (defknown safe-fdefn-fun (fdefn) function ())
 
 (defknown %simple-fun-type (function) t (flushable))
 
-#+(or x86 x86-64) (defknown sb-vm::%closure-callee (function) fixnum (flushable))
+#+(or x86 x86-64 arm64) (defknown sb-vm::%closure-callee (function) fixnum (flushable))
 (defknown %closure-fun (function) function (flushable))
 
 (defknown %closure-index-ref (function index) t
   (flushable))
+(defknown %closure-index-set (function index t) (values) ())
 
 ;; T argument is for the 'fun' slot.
 (defknown sb-vm::%alloc-closure (index t) function (flushable))
@@ -574,7 +614,8 @@
   ())
 
 (defknown %funcallable-instance-info (function index) t (flushable))
-(defknown %set-funcallable-instance-info (function index t) t ())
+(defknown (setf %funcallable-instance-info) (t function index) t ())
+(defknown %set-funcallable-instance-info (function index t) (values) ())
 
 #+sb-fasteval
 (defknown sb-interpreter:fun-proto-fn (interpreted-function)
@@ -592,6 +633,9 @@
 (defknown %single-float (real) single-float (movable foldable))
 (defknown %double-float (real) double-float (movable foldable))
 
+(defknown bignum-to-float (bignum symbol) float (movable foldable))
+(defknown sb-kernel::float-ratio (ratio symbol) float (movable foldable))
+
 (defknown make-single-float ((signed-byte 32)) single-float
   (movable flushable))
 
