@@ -1,16 +1,5 @@
 (in-package "SB-X86-64-ASM")
 
-(defun get-avx2 (number)
-  (svref (load-time-value
-          (coerce (loop for i from 0 below 16
-                        collect (!make-reg (!make-avx2-id i)))
-                  'vector)
-          t)
-         number))
-
-(defun is-avx2-id-p (reg-id)
-  (= (ldb (byte 3 0) reg-id) 3))
-
 (define-arg-type ymmreg
   :prefilter #'prefilter-reg-r
   :printer #'print-ymmreg)
@@ -228,12 +217,12 @@
                0
                1))
          (xmm-size (r)
-           (cond ((is-avx2-id-p (reg-id r))
+           (cond ((is-ymm-id-p (reg-id r))
                   1)
                  ((xmm-register-p r)
                   0))))
     (let ((l (cond (l)
-                   (reg
+                   ((xmm-register-p reg)
                     (xmm-size reg))
                    ((xmm-register-p thing)
                     (xmm-size thing))
@@ -584,7 +573,7 @@
 (macrolet ((def (name prefix)
              `(define-instruction ,name (segment dst src pattern)
                 ,@(avx2-inst-printer-list
-                   'ymm-ymm/mem prefix #x70
+                   'ymm-ymm/mem-imm prefix #x70
                    :printer '(:name :tab reg ", " reg/mem ", " imm))
                 (:emitter
                  (emit-avx2-inst segment dst src ,prefix #x70
@@ -764,14 +753,12 @@
                 (:emitter
                  (cond ((and (xmm-register-p dst)
                              (ea-p src))
-                        (emit-avx2-inst segment dst src ,prefix #x10 :l 0))
+                        (emit-avx2-inst segment src dst ,prefix #x10 :l 0))
                        ((xmm-register-p dst)
-                        (emit-avx2-inst segment dst src2 ,prefix #x10 :vvvv src
-                                                                      :l 0))
+                        (emit-avx2-inst segment src2 dst ,prefix #x10 :vvvv src :l 0))
                        (t
                         (aver (xmm-register-p src))
-                        (emit-avx2-inst segment src dst ,prefix #x11
-                                        :l 0)))))))
+                        (emit-avx2-inst segment dst src ,prefix #x11 :l 0)))))))
   (def vmovsd #xf2)
   (def vmovss #xf3))
 
@@ -1121,11 +1108,11 @@
                                  :w ,w
                                  :l ,(ecase sizing
                                        ((xmm/ymm-vmx/y xmm/ymm-vmx)
-                                        `(if (is-avx2-id-p (reg-id dst))
+                                        `(if (is-ymm-id-p (reg-id dst))
                                              1
                                              0))
                                        (xmm-vmx/y
-                                        `(if (eq (sc-name (tn-sc (ea-index vm))) 'avx2-reg)
+                                        `(if (eq (sc-name (tn-sc (ea-index vm))) 'ymm-reg)
                                              1
                                              0)))
                                  :vm t)))))
