@@ -19,6 +19,7 @@
 #include "genesis/avlnode.h"
 #include "genesis/sap.h"
 #include "genesis/thread-instance.h"
+#include "print.h"
 
 #include <stdlib.h>
 #include <stdio.h>
@@ -209,7 +210,7 @@ static char* NO_SANITIZE_MEMORY deduce_thread_pc(struct thread* th, void** addr)
     uword_t* fp = __builtin_frame_address(0);
     char* return_pc = 0;
 
-    if (th != arch_os_get_current_thread()) {
+    if (th != get_sb_vm_thread()) {
         int i = fixnum_value(read_TLS(FREE_INTERRUPT_CONTEXT_INDEX,th));
         os_context_t *c = nth_interrupt_context(i-1, th);
         fp = (uword_t*)*os_context_register_addr(c,reg_FP);
@@ -243,7 +244,7 @@ deduce_thread(void (*context_scanner)(), uword_t pointer, char** pc)
     for_each_thread(th) {
         void **esp=(void **)-1;
         sword_t i,free;
-        if (th == arch_os_get_current_thread())
+        if (th == get_sb_vm_thread())
             esp = (void **)((void *)&pointer);
         else {
             void **esp1;
@@ -350,7 +351,7 @@ static lispobj examine_threads(struct hopscotch_table* targets,
                 *thread_pc = 0;
                 // Scan just the current thread's stack
                 // (We don't know where the other stack pointers are)
-                th = arch_os_get_current_thread();
+                th = get_sb_vm_thread();
                 void **esp = __builtin_frame_address(0);
                 void **where;
                 for (where = ((void **)th->control_stack_end)-1; where >= esp;  --where)
@@ -360,7 +361,7 @@ static lispobj examine_threads(struct hopscotch_table* targets,
                         break;
                     }
             }
-            return pin;
+            if (*root_thread) return pin;
         }
 #endif
     *root_kind = HEAP;
@@ -430,20 +431,20 @@ static inline lispobj decode_pointer(uint32_t encoding)
 
 static void maybe_show_object_name(lispobj obj, FILE* stream)
 {
-    extern void safely_show_lstring(lispobj* string, int quotes, FILE *s);
     lispobj package, package_name;
     if (lowtag_of(obj)==OTHER_POINTER_LOWTAG)
         switch(widetag_of(native_pointer(obj))) {
         case SYMBOL_WIDETAG:
             putc(',', stream);
-            if ((package = SYMBOL(obj)->package) == NIL) {
+            package = symbol_package(SYMBOL(obj));
+            if (package == NIL) {
                 fprintf(stream, "#:");
             } else {
                 package_name = ((struct package*)native_pointer(package))->_name;
-                safely_show_lstring(native_pointer(package_name), 0, stream);
+                safely_show_lstring(VECTOR(package_name), 0, stream);
                 fputs("::", stream);
             }
-            safely_show_lstring(native_pointer(SYMBOL(obj)->name), 0, stream);
+            safely_show_lstring(symbol_name(SYMBOL(obj)), 0, stream);
             break;
         }
 }
@@ -463,7 +464,7 @@ static boolean root_p(lispobj ptr, int criterion)
 static lispobj mkcons(lispobj car, lispobj cdr)
 {
     struct cons *cons = (struct cons*)
-        gc_general_alloc(sizeof(struct cons), BOXED_PAGE_FLAG);
+        gc_general_alloc(cons_region, sizeof(struct cons), PAGE_TYPE_CONS);
     cons->car = car;
     cons->cdr = cdr;
     return make_lispobj(cons, LIST_POINTER_LOWTAG);
@@ -474,7 +475,7 @@ static lispobj liststar3(lispobj x, lispobj y, lispobj z) {
 static lispobj make_sap(char* value)
 {
     struct sap *sap = (struct sap*)
-        gc_general_alloc(sizeof(struct sap), BOXED_PAGE_FLAG);
+        gc_general_alloc(unboxed_region, sizeof(struct sap), PAGE_TYPE_UNBOXED);
     sap->header = (1<<N_WIDETAG_BITS) | SAP_WIDETAG;
     sap->pointer = value;
     return make_lispobj(sap, OTHER_POINTER_LOWTAG);
@@ -703,7 +704,7 @@ static boolean ignorep(lispobj* base_ptr,
                        lispobj ignored_objects)
 {
     int i;
-    for (i = fixnum_value(VECTOR(ignored_objects)->length)-1; i >= 0; --i)
+    for (i = vector_len(VECTOR(ignored_objects))-1; i >= 0; --i)
       if (native_pointer(VECTOR(ignored_objects)->data[i]) == base_ptr)
           return 1;
     return 0;
@@ -764,13 +765,12 @@ static uword_t build_refs(lispobj* where, lispobj* end,
             // a path from the root only if it does not involve whichever
             // object is definitely weak. This fails on weak key-OR-value
             // tables since we can't decide whether to allow the entry.
-            // Disregard the address-sensitivity bit when checking whether
-            // this is a weak hash-table vector.
-            if ((vector_subtype(*where) & ~subtype_VectorAddrHashing)
-                == subtype_VectorHashing + subtype_VectorWeak) {
+            if (vector_flagp(*where, VectorWeak)) { // if any kind of weak vector
+                // If not a hashtable, just skip it.
+                if (!vector_flagp(*where, VectorHashing)) continue;
                 lispobj* data = where + 2;
-                int kv_vector_length = fixnum_value(where[1]);
-                lispobj lhash_table = data[kv_vector_length-1];
+                int kv_vector_len = vector_len((struct vector*)where);
+                lispobj lhash_table = data[kv_vector_len-1];
                 gc_assert(instancep(lhash_table));
                 struct hash_table* hash_table =
                   (struct hash_table *)native_pointer(lhash_table);
@@ -785,8 +785,6 @@ static uword_t build_refs(lispobj* where, lispobj* end,
                     continue;
                 }
             }
-            if (is_vector_subtype(*where, VectorWeak))
-                continue;
             break;
         default:
             if (!(other_immediate_lowtag_p(widetag) && LOWTAG_FOR_WIDETAG(widetag)))
@@ -837,6 +835,7 @@ static uword_t build_refs(lispobj* where, lispobj* end,
 static void scan_spaces(struct scan_state* ss)
 {
     struct scan_state old = *ss;
+    build_refs((lispobj*)NIL_SYMBOL_SLOTS_START, (lispobj*)NIL_SYMBOL_SLOTS_END, ss);
     build_refs((lispobj*)STATIC_SPACE_OBJECTS_START, static_space_free_pointer, ss);
     show_tally(old, ss);
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
@@ -938,6 +937,7 @@ static boolean finding_leaf_p(lispobj weak_pointers)
     return 0; // this is the expected (and optimal) case
 }
 
+extern void gc_close_all_regions();
 /* Find any shortest path from a thread or tenured object
  * to each of the specified objects.
  */
@@ -969,7 +969,8 @@ static int trace_paths(void (*context_scanner)(),
     hopscotch_create(&targets, HASH_FUNCTION, 0, 32, 0);
     i = 0;
     do {
-        gc_assert(make_fixnum(i) <= VECTOR(paths)->length);
+        // Oh dear, is this really supposed to be '<=' (vs '<') ?
+        gc_assert(i <= vector_len(VECTOR(paths)));
         lispobj car = CONS(weak_pointers)->car;
         lispobj value = ((struct weak_pointer*)native_pointer(car))->value;
         weak_pointers = CONS(weak_pointers)->cdr;
@@ -982,11 +983,13 @@ static int trace_paths(void (*context_scanner)(),
                                   &targets, &visited,
                                   inverted_heap, &scratchpad,
                                   n_pins, pins, context_scanner, criterion);
-            if ((VECTOR(paths)->data[i] = path) != 0) ++n_found;
+            lispobj* elt = VECTOR(paths)->data + i;
+            notice_pointer_store(elt);
+            if ((*elt = path) != 0) ++n_found;
         }
         ++i;
     } while (weak_pointers != NIL);
-    ensure_region_closed(&boxed_region, BOXED_PAGE_FLAG);
+    gc_close_all_regions();
     os_invalidate(scratchpad.base, scratchpad.end-scratchpad.base);
 #if TRACEROOT_USE_ABSL_HASHMAP
     absl_hashmap_destroy(inverted_heap);
@@ -1046,15 +1049,6 @@ int gc_prove_liveness(void(*context_scanner)(),
         CONS(output)->car = make_fixnum(0);
         return 0;
     }
-    // Put back lowtags on pinned objects, since wipe_nonpinned_words() removed
-    // them. But first test whether lowtags were already repaired
-    // in case prove_liveness() is called after gc_prove_liveness().
-    if (n_pins>0 && !is_lisp_pointer(pins[0])) {
-        int i;
-        for(i=n_pins-1; i>=0; --i) {
-            pins[i] = compute_lispobj((lispobj*)pins[i]);
-        }
-    }
     n_paths = trace_paths(context_scanner, input, paths, ignore,
                           n_pins, (lispobj*)pins, criterion);
     CONS(output)->car = make_fixnum(n_paths);
@@ -1067,6 +1061,7 @@ int gc_prove_liveness(void(*context_scanner)(),
 int prove_liveness(lispobj objects, int criterion)
 {
     extern struct hopscotch_table pinned_objects;
-    extern int gc_n_stack_pins;
-    return gc_prove_liveness(0, objects, gc_n_stack_pins, pinned_objects.keys, criterion);
+    extern int gc_pin_count;
+    extern lispobj* gc_filtered_pins;
+    return gc_prove_liveness(0, objects, gc_pin_count, gc_filtered_pins, criterion);
 }
